import numpy
import random
from math import exp, log, sqrt
import matplotlib.pyplot as plt

random.seed(1234)

dim = 5
sg = 0.01
n = 5
v = [numpy.zeros(dim) for i in range(2**n)]
for i, vi in enumerate(v):
    for j in range(n):
        vi[j] = (-1.0)**(i // 2**j)


# x1 = numpy.zeros(dim)
# x2 = numpy.zeros(dim)
# x1[0] = x1[1] = -1.0
# x2[0] = x2[1] = +1.0

def lnf(x):
    d = [x - vi for vi in v]
    p = dim + 2
    try:
        return log(sum(sg**p/(sg**p + numpy.dot(di,di)**(0.5*p)) 
                       for di in d))
    except ValueError:
        return -numpy.inf

class Agent:
    def __init__(self, x, lnf, peers):
        self.x = x
        self.lnf = lnf
        self.peers = peers

def dsq(x,y):
    d = y-x
    return numpy.dot(d,d)

nagents = 500
npeers = 50
ncandidates = 200

agents = []
for i in range(nagents):
    x = numpy.array([random.gauss(0.0, 2.0) for j in range(dim)])
    peers = random.sample(range(0, i-1)+range(i+1, nagents), npeers)
    agents.append(Agent(x, lnf(x), peers))

nsweeps = 10
reject = 0
for sweep in range(50):
    if sweep % nsweeps == 0:
        crosses = 0
        for a in agents:
            for p in a.peers:
                if dsq(a.x, agents[p].x) >= 1.0:
                    crosses = crosses + 1
        acceptance = (1.0 - reject/(10.0*nagents))
        reject = 0
        print crosses, acceptance*100.0

    # if sweep % 100 == 0:
    #     plt.scatter([a.x[0] for a in agents], [a.x[1] for a in agents])
    #     plt.xlim(-5.0, 5.0)
    #     plt.ylim(-5.0, 5.0)
    #     plt.show()

    newagents = []
    for k, ak in enumerate(agents):
        candidates = set(ak.peers)
        candidates.update(random.sample(xrange(nagents), ncandidates))
        candidates.discard(k)
        candidates = [(dsq(ak.x, agents[c].x), c) for c in candidates]
        candidates.sort()
        ak.peers = [candidates[i][1] for i in range(npeers)]

        if random.uniform(0.0, 1.0) <= acceptance:
            j = random.randrange(len(agents)-1)
            aj = agents[j] if j >= k else agents[j+1]
        else:
            aj = agents[random.choice(ak.peers)]
        z = 0.5*(1.0 + random.uniform(0.0, 1.0))**2
        y = aj.x + z*(ak.x - aj.x)
        lnfy = lnf(y)
        lnpaccept = min((dim-1.0)*log(z) + lnfy - ak.lnf, 0.0)
        if random.uniform(0.0, 1.0) <= exp(lnpaccept):
            newagents.append(Agent(y, lnfy, ak.peers))
        else:
            newagents.append(ak)
            reject = reject + 1
    agents = newagents

for vi in v:
    print sum(1 if dsq(vi, a.x) <= 1.0 else 0 for a in agents)

        
    
